/*
  TrueType drawing for mnw
  Copyright (c) 1998 FURUKAWA Yo.
  $Id: mnwtt.c,v 1.6 1999/03/09 00:20:36 furukawa Exp $
  */

/* NOTE: 1 face, 16 instances are available */

#define MNW_INLINLINING_FOR_SPEED
#define MNW_USE_SHORTCUT
#include "arch.h"
#include "mnw.h"

#ifdef MNW_USE_TRUETYPE
#include "freetype.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* -- file local macros */

#define MNWTT_MAX_FONTID	16

/* -- file local types */

typedef struct _mnw_tt_font{
    /* per instance */
    TT_Instance instance;
    TT_Instance_Metrics instance_metrics;
    int descent;
    bool gray_render;
    int width;
    int height;
    /* raster buffer */
    TT_Raster_Map result_map;
    TT_Raster_Map gray_map;
} mnwTTFont;

/* -- file local variables */

/* share */
static TT_Engine gTT_Engine;
static TT_Face gTT_Face;
static TT_Face_Properties gTT_Properties;
static TT_Glyph gTT_Glyph;
static TT_CharMap gTT_CharMap;

/* per instance */
static mnwTTFont gTT_Font[MNWTT_MAX_FONTID];
static int gTT_FontNum;

/* enum */
unsigned char gTT_VirtualPalette[5] = { 0, 1, 2, 3, 4 };
unsigned char gTT_BoundedPalette[8] = { 0, 1, 2, 3, 4, 4, 4, 4 };


/* -- file local methods */

/* -- methods */

/* -- mnwTT --------------------------------- methods */

bool
mnwTT_Initialize(char *file_name)
{
    TT_Error error;
    bool charmap_found;
    int i;

    /* init engine */
    if ((error = TT_Init_FreeType(&gTT_Engine))){
	mnwDPrintf("Error while initializing engine, code = %d.", error);
	return false;
    }

    /* open face */
    error = TT_Open_Face(gTT_Engine, file_name, &gTT_Face);
    if (error == TT_Err_Could_Not_Open_File){
	mnwDPrintf("Could not find/open %s.", file_name);
	return false;
    }else if (error){
	mnwDPrintf("Error while opening %s, code = %x.", file_name, error);
	return false;
    }
    TT_Get_Face_Properties(gTT_Face, &gTT_Properties);
    mnwDPrintf("%s: open success %d glyphs.",
	       file_name, gTT_Properties.num_Glyphs);

    /* create glyph */
    error = TT_New_Glyph(gTT_Face, &gTT_Glyph);
    if (error){
	mnwDPrintf("Could not create glyph container.");
	return false;
    }

    /* init char map */
    charmap_found = false;
    for (i = 0; i < gTT_Properties.num_CharMaps; i++){
	TT_UShort platform, encoding;
	TT_Get_CharMap_ID(gTT_Face, i, &platform, &encoding);
	if (platform == 3 && encoding == 1){
	    charmap_found = true;
	    TT_Get_CharMap(gTT_Face, i, &gTT_CharMap);
	}
    }
    if (!charmap_found){
	mnwDPrintf("Could not found Windows Unicode character mapping.");
	return false;
    }

    return true;
}

bool
mnwTT_Finalize(void)
{
    TT_Close_Face(gTT_Face);
    TT_Done_FreeType(gTT_Engine);	      /* release ALL objects */
    return true;
}

/* -- mnwTTFont --------------------------------- methods */

static inline void
mnwTTFont_ClearResultMap(mnwTTFont *self)
{
    memset(self->result_map.bitmap, 0, self->result_map.size);
}

static inline void
mnwTTFont_ClearGrayMap(mnwTTFont *self)
{
    memset(self->gray_map.bitmap, 0, self->gray_map.size);
}

static bool
mnwTTFont_InitResultMap(mnwTTFont *self)
{
    self->result_map.rows = self->height;
    self->result_map.width = self->width + 8;
    self->result_map.flow = TT_Flow_Down;
    if (self->gray_render){
	self->result_map.cols = (self->result_map.width+3) & -4;
    }else{
	self->result_map.cols = (self->result_map.width+7) >> 3;
    }
    self->result_map.size = (long)self->result_map.cols * self->result_map.rows;
    
    if (self->result_map.bitmap){
	free(self->result_map.bitmap);
    }
    self->result_map.bitmap = malloc((int)self->result_map.size);
    if (!self->result_map.bitmap){
	return false;
    }
      
    mnwTTFont_ClearResultMap(self);
    return true;
}

static bool
mnwTTFont_InitGrayMap(mnwTTFont *self)
{
    if (self->gray_map.bitmap){
	free(self->gray_map.bitmap);
    }
    
    self->gray_map.rows = self->height + 32;
    self->gray_map.width = self->width + 32;
    self->gray_map.cols = (self->gray_map.width+3) & -4;
    self->gray_map.flow = TT_Flow_Up;
    self->gray_map.size = self->gray_map.rows * self->gray_map.cols;

    self->gray_map.bitmap = malloc((int)self->gray_map.size);
    if (self->gray_map.bitmap){
	return false;
    }
    
    mnwTTFont_ClearGrayMap(self);
    return true;
}

int
mnwTTFont_Create(int point_size, int res, bool gray_render)
{
    int upm;
    TT_Error error;
    mnwTTFont *self;

    /* allocate mnwTTFont */
    if (gTT_FontNum >= MNWTT_MAX_FONTID){
	return false;
    }
    self = &gTT_Font[gTT_FontNum++];

    /* create instance */
    error = TT_New_Instance(gTT_Face, &(self->instance));
    if (error){
	mnwDPrintf("Could not create instance.");
	return false;
    }
    error = TT_Set_Instance_Resolutions(self->instance, res, res);
    if (error){
	mnwDPrintf("Could not set device resolutions.");
	return false;
    }
    if (gray_render){
	TT_Set_Raster_Gray_Palette(gTT_Engine, gTT_VirtualPalette);
    }

    /* reset scale */
    if ((error = TT_Set_Instance_CharSize(self->instance, point_size*64))){
	mnwDPrintf("Error = %d.", (int)error);
	mnwDPrintf("Could not reset instance.");
	return false;
    }
    TT_Get_Instance_Metrics(self->instance, &(self->instance_metrics));

    /* set property */
    self->gray_render = gray_render;
    self->width = self->instance_metrics.x_ppem;
    self->height = self->instance_metrics.y_ppem;
    upm = gTT_Properties.header->Units_Per_EM;
    self->descent = (gTT_Properties.horizontal->Descender * self->height)/upm;

    /* init bitmap */
    mnwTTFont_InitResultMap(self);
    if (gray_render){
	mnwTTFont_InitGrayMap(self);
    }

    return gTT_FontNum - 1;		      /* font id */
}

static inline TT_Error
mnwTTFont_RenderBitmapGlyph(mnwTTFont *self, TT_Glyph glyph,
			    int x_offset, int y_offset)
{
    return TT_Get_Glyph_Bitmap(glyph, &(self->result_map),
			       (long)x_offset*64, (long)y_offset*64);
}

static TT_Error
mnwTTFont_RenderPixmapGlyph(mnwTTFont *self, TT_Glyph glyph,
			    int x_offset, int y_offset)
{
    TT_Error error;
    TT_Glyph_Metrics metrics;
    TT_F26Dot6 x, y, xmin, ymin, xmax, ymax;
    int ioff, iread;
    char *off, *read, *_off, *_read;
    
    /* grid-fitg the bounding box */
    TT_Get_Glyph_Metrics(glyph, &metrics);
    xmin = metrics.bbox.xMin & -64;
    ymin = metrics.bbox.yMin & -64;
    xmax = (metrics.bbox.xMax + 63) & -64;
    ymax = (metrics.bbox.yMax + 63) & -64;
    
    mnwTTFont_ClearGrayMap(self);
    error = TT_Get_Glyph_Pixmap(glyph, &(self->gray_map), -xmin, -ymin);
    if (error)
	return error;

    xmin = (xmin >> 6) + x_offset;
    ymin = (ymin >> 6) + y_offset;
    xmax = (xmax >> 6) + x_offset;
    ymax = (ymax >> 6) + y_offset;

    if (xmin >= (int)self->result_map.width
	|| ymin >= (int)self->result_map.rows
	|| xmax < 0
	|| ymax < 0)
	return TT_Err_Ok;  /* nothing to do */
    
    if (xmax - xmin + 1 > self->gray_map.width)
	xmax = xmin + self->gray_map.width - 1;
    if (ymax - ymin + 1 > self->gray_map.rows)
	ymax = ymin + self->gray_map.rows - 1;

    iread = 0;
    if (ymin < 0){
	iread -= ymin * self->gray_map.cols;
	ioff   = 0;
	ymin   = 0;
    }else{
	ioff = ymin * self->result_map.cols;
    }

    if (ymax >= self->result_map.rows)
	ymax = self->result_map.rows-1;

    if (xmin < 0){
	iread -= xmin;
	xmin   = 0;
    }else{
	ioff += xmin;
    }

    if (xmax >= self->result_map.width)
	xmax = self->result_map.width - 1;

    _read = (char*)self->gray_map.bitmap + iread;
    _off  = (char*)self->result_map.bitmap + ioff;

    for (y = ymin; y <= ymax; y++){
	read = _read;
	off  = _off;

	for (x = xmin; x <= xmax; x++){
	    *off = *read; /* *off = gTT_BoundedPalette[*off | *read]; */
	    off++;
	    read++;
	}
	_read += self->gray_map.cols;
	_off  += self->result_map.cols;
    }

    return TT_Err_Ok;
}

static void
mnwTTFont_DrawBitmap(mnwTTFont *self,
		     int left, int top, int width, int height,
		     int direction)
{
    int i, row, col, x, y;
    unsigned char patt;

    switch (direction){
    case kDIR_NORTH:
	for (row = 0; row < height; row++){
	    i = row * self->result_map.cols;
	    x = left;
	    y = top + row;
	    for (col = 0; col < self->result_map.cols; col++){
		patt = ((unsigned char *)self->result_map.bitmap)[i++];
		if (patt & (0x01<<7)) mnwDrawPixel(x + 0, y);
		if (patt & (0x01<<6)) mnwDrawPixel(x + 1, y);
		if (patt & (0x01<<5)) mnwDrawPixel(x + 2, y);
		if (patt & (0x01<<4)) mnwDrawPixel(x + 3, y);
		if (patt & (0x01<<3)) mnwDrawPixel(x + 4, y);
		if (patt & (0x01<<2)) mnwDrawPixel(x + 5, y);
		if (patt & (0x01<<1)) mnwDrawPixel(x + 6, y);
		if (patt & (0x01<<0)) mnwDrawPixel(x + 7, y);
		x += 8;
	    }
	}
	break;
    case kDIR_SOUTH:
	for (row = 0; row < height; row++){
	    i = row * self->result_map.cols;
	    x = left + width;
	    y = top + height - row;
	    for (col = 0; col < self->result_map.cols; col++){
		patt = ((unsigned char *)self->result_map.bitmap)[i++];
		if (patt & (0x01<<7)) mnwDrawPixel(x + 7, y);
		if (patt & (0x01<<6)) mnwDrawPixel(x + 6, y);
		if (patt & (0x01<<5)) mnwDrawPixel(x + 5, y);
		if (patt & (0x01<<4)) mnwDrawPixel(x + 4, y);
		if (patt & (0x01<<3)) mnwDrawPixel(x + 3, y);
		if (patt & (0x01<<2)) mnwDrawPixel(x + 2, y);
		if (patt & (0x01<<1)) mnwDrawPixel(x + 1, y);
		if (patt & (0x01<<0)) mnwDrawPixel(x + 0, y);
		x -= 8;
	    }
	}
	break;
    case kDIR_EAST:
	for (row = 0; row < height; row++){
	    i = row * self->result_map.cols;
	    x = left + width - row;
	    y = top;
	    for (col = 0; col < self->result_map.cols; col++){
		patt = ((unsigned char *)self->result_map.bitmap)[i++];
		if (patt & (0x01<<7)) mnwDrawPixel(x, y + 0);
		if (patt & (0x01<<6)) mnwDrawPixel(x, y + 1);
		if (patt & (0x01<<5)) mnwDrawPixel(x, y + 2);
		if (patt & (0x01<<4)) mnwDrawPixel(x, y + 3);
		if (patt & (0x01<<3)) mnwDrawPixel(x, y + 4);
		if (patt & (0x01<<2)) mnwDrawPixel(x, y + 5);
		if (patt & (0x01<<1)) mnwDrawPixel(x, y + 6);
		if (patt & (0x01<<0)) mnwDrawPixel(x, y + 7);
		y += 8;
	    }
	}
	break;
    case kDIR_WEST:
	for (row = 0; row < height; row++){
	    i = row * self->result_map.cols;
	    x = left + row;
	    y = top + height;
	    for (col = 0; col < self->result_map.cols; col++){
		patt = ((unsigned char *)self->result_map.bitmap)[i++];
		if (patt & (0x01<<7)) mnwDrawPixel(x, y + 7);
		if (patt & (0x01<<6)) mnwDrawPixel(x, y + 6);
		if (patt & (0x01<<5)) mnwDrawPixel(x, y + 5);
		if (patt & (0x01<<4)) mnwDrawPixel(x, y + 4);
		if (patt & (0x01<<3)) mnwDrawPixel(x, y + 3);
		if (patt & (0x01<<2)) mnwDrawPixel(x, y + 2);
		if (patt & (0x01<<1)) mnwDrawPixel(x, y + 1);
		if (patt & (0x01<<0)) mnwDrawPixel(x, y + 0);
		y -= 8;
	    }
	}
	break;
    }
}

#define kColor_None 256

static int gMNWTTPalette[4][5] = 
{
    { kColor_None, kColor_DkGray, kColor_LtGray, kColor_LtGray, kColor_White, },
    { kColor_None, kColor_DkGray, kColor_DkGray, kColor_LtGray, kColor_LtGray, },
    { kColor_None, kColor_LtGray, kColor_LtGray, kColor_DkGray, kColor_DkGray, },
    { kColor_None, kColor_LtGray, kColor_DkGray, kColor_DkGray, kColor_Black, },
};

static void
mnwTTFont_DrawPixmap(mnwTTFont *self,
		     int left, int top, int width, int height,
		     int direction)
{
    int i, row, col, x, y, color, draw_color;
    unsigned char patt;

    color = gDefaultGraphics->fore_color;

    switch (direction){
    case kDIR_NORTH:
	for (row = 0; row < height; row++){
	    i = row * self->result_map.cols;
	    x = left;
	    y = top + height - row;
	    for (col = 0; col < width; col++){
		patt = ((unsigned char *)self->result_map.bitmap)[i++];
		draw_color = gMNWTTPalette[color][patt];
		if (draw_color != kColor_None){
		    mnwSetForeColor(draw_color);
		    mnwDrawPixel(x, y);
		}
		x++;
	    }
	}
	break;
    case kDIR_SOUTH:
	for (row = 0; row < height; row++){
	    i = row * self->result_map.cols;
	    x = left + width;
	    y = top + row;
	    for (col = 0; col < width; col++){
		patt = ((unsigned char *)self->result_map.bitmap)[i++];
		draw_color = gMNWTTPalette[color][patt];
		if (draw_color != kColor_None){
		    mnwSetForeColor(draw_color);
		    mnwDrawPixel(x, y);
		}
		x--;
	    }
	}
	break;
    case kDIR_EAST:
	for (row = 0; row < height; row++){
	    i = row * self->result_map.cols;
	    x = left + row;
	    y = top;
	    for (col = 0; col < width; col++){
		patt = ((unsigned char *)self->result_map.bitmap)[i++];
		draw_color = gMNWTTPalette[color][patt];
		if (draw_color != kColor_None){
		    mnwSetForeColor(draw_color);
		    mnwDrawPixel(x, y);
		}
		y++;
	    }
	}
	break;
    case kDIR_WEST:
	for (row = 0; row < height; row++){
	    i = row * self->result_map.cols;
	    x = left + width - row;
	    y = top + height;
	    for (col = 0; col < width; col++){
		patt = ((unsigned char *)self->result_map.bitmap)[i++];
		draw_color = gMNWTTPalette[color][patt];
		if (draw_color != kColor_None){
		    mnwSetForeColor(draw_color);
		    mnwDrawPixel(x, y);
		}
		y--;
	    }
	}
	break;
    }

    mnwSetForeColor(color);		      /* recover */
}

static bool
mnwTTFont_DrawByIndex(mnwTTFont *self,
		      int index, int x, int y, int width, int height,
		      int direction)
{
    TT_Error error;

    error = TT_Load_Glyph(self->instance, gTT_Glyph,
			  index, TTLOAD_SCALE_GLYPH|TTLOAD_HINT_GLYPH);
    if (error)
	return false;

    mnwTTFont_ClearResultMap(self);
    if (!self->gray_render){
	mnwTTFont_RenderBitmapGlyph(self, gTT_Glyph, 0, -self->descent);
	mnwTTFont_DrawBitmap(self, x, y, width, height, direction);
    }else{
	mnwTTFont_RenderPixmapGlyph(self, gTT_Glyph, 0, -self->descent);
	mnwTTFont_DrawPixmap(self, x, y, width, height, direction);
    }
    return true;
}

/* --- export */

void
mnwTT_DrawText(int id, int x, int y, char *text, int len)
{
    char *p;
    int remain;
    int index;
    unsigned short code;
    mnwTTFont *self = &(gTT_Font[id]);

    remain = len;
    p = text;
    while (*p && remain > 0){
	if (*p & 0x80) {
	    /* EUC 2byte char */
	    if (remain < 2){
		return;
	    }
	    if ((p[0]&0xff) == 0x8E){
		/* EUC x0201 kana */
		code = p[1]&0xff;
		index = TT_Char_Index(gTT_CharMap, conv_JISX0201toUCS2(code));
		mnwTTFont_DrawByIndex(self, index, x, y,
				      self->width / 2,
				      self->height,
				      kDIR_NORTH);
		x += self->width / 2;
	    }else{
		/* EUC x0208 */
		code = ((p[0]&0x7F)<<8)|(p[1]&0x7F);
		index = TT_Char_Index(gTT_CharMap, conv_JISX0208toUCS2(code));
		mnwTTFont_DrawByIndex(self, index, x, y,
				      self->width,
				      self->height,
				      kDIR_NORTH);
		x += self->width;
	    }
	    p += 2;
	    remain -= 2;
	}else{
	    /* 1byte char */
	    index = TT_Char_Index(gTT_CharMap, conv_JISX0201toUCS2(*p & 0xff));
	    mnwTTFont_DrawByIndex(self, index, x, y,
				  self->width / 2,
				  self->height,
				  kDIR_NORTH);
	    x += self->width / 2;
	    p++;
	    remain--;
	}
    }
    mnwFlush();
}

void
mnwTT_DrawFontRotated(int id, int x, int y, int code, int direction)
{
    int index;
    mnwTTFont *self = &(gTT_Font[id]);

    if (code > 0x100) {
	/* EUC 2byte char */
	if ((code&0xff00) == 0x8E00){
	    /* EUC x0201 kana */
	    code = code & 0xff;
	    index = TT_Char_Index(gTT_CharMap, conv_JISX0201toUCS2(code));
	    mnwTTFont_DrawByIndex(self, index, x, y,
				  self->width / 2,
				  self->height,
				  direction);
	    x += self->width / 2;
	}else{
	    /* EUC x0208 */
	    code = code & 0x7f7f;
	    index = TT_Char_Index(gTT_CharMap, conv_JISX0208toUCS2(code));
	    mnwTTFont_DrawByIndex(self, index, x, y,
				  self->width,
				  self->height,
				  direction);
	    x += self->width;
	}
    }else{
	/* 1byte char */
	index = TT_Char_Index(gTT_CharMap, conv_JISX0201toUCS2(code & 0xff));
	mnwTTFont_DrawByIndex(self, index, x, y,
			      self->width / 2,
			      self->height,
			      direction);
	x += self->width / 2;
    }
    mnwFlush();
}

int
mnwTT_GetFontWidth(int id)
{
    mnwTTFont *self = &(gTT_Font[id]);
    return self->width;
}

int
mnwTT_GetFontHeight(int id)
{
    mnwTTFont *self = &(gTT_Font[id]);
    return self->height;
}

